import java.io.File;
import java.nio.ByteBuffer;
import java.util.Arrays;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;


public class xmaextract {

	public static final byte[] XMA2_ID = {0x58, 0x4D, 0x41, 0x32};
	public static final byte[] FSB4_ID = {0x46, 0x53, 0x42, 0x34};
	public static void main(String[] args) throws IOException
	{
		if (args.length != 1)
		{
			printUsage();
			return;
		}
		File xmaFile = new File(args[0]);
		System.out.println("Reading xma File... "+xmaFile.length()+" bytes.");
		FileInputStream xmaStream = new FileInputStream(xmaFile);
		ByteBuffer xmaBuffer = ByteBuffer.allocate( (int) xmaFile.length());
		xmaStream.read(xmaBuffer.array());
		xmaStream.close();
		byte[] fileHeader = new byte[4];
		System.arraycopy(xmaBuffer.array(), 0, fileHeader, 0, 4);
		if (Arrays.equals(fileHeader, FSB4_ID))
		{
			//let's process it like mp3extract.
			//Find first packet.
			//When First packet found. Start interleaving.
			int sizeHead = 0;
			for (int i=3; i>-1; i--)
			{
				int numadd = xmaBuffer.array()[8+i];
				if (numadd<0) numadd += 256; //WE need to convert byte to int. Since java uses signed bytes
				//we need to add 256 to negative numbers. I guess we could unary minus it (flip bits add 1)
				//But using human math rather than binary math is easier.
				sizeHead += numadd*((int) (Math.pow(16, i*2)));
			}
			int packetOffset = 48+sizeHead;
			xmaBuffer.position(packetOffset);
			int numStreams = getStreams(xmaBuffer, packetOffset);
			for (int i=0; i<numStreams; i++)
			{
				System.out.println("Extracting Channel "+(i+1));
				boolean eofReached = false;
				int skipCount = 0;
				String outName = xmaFile.getAbsolutePath().concat("_channel_"+(i+1)+".xma");
				FileOutputStream xmaOutStream = new FileOutputStream(outName);
				while(!eofReached)
				{
					//Extract each one.
					//Read in a packet.
					byte[] nextPacket = new byte[2048];
					if (xmaBuffer.remaining() < 2048)
					{
						break;
					}
					xmaBuffer.get(nextPacket);
					if (skipCount == 0)
					{
						xmaOutStream.write(nextPacket);
						//reset skipcount
						int byteSkipCount = xmaBuffer.array()[xmaBuffer.position()+3-2048];
						if (byteSkipCount < 0) byteSkipCount+=256;
						skipCount = byteSkipCount;
					}
					else
						skipCount--;
					if (!xmaBuffer.hasRemaining())
					{
						eofReached = true;
					}
					
				}
				xmaOutStream.close();
				xmaBuffer.position(packetOffset+2048*(i+1));
				
			}
			//File number of channels.
		}
		/*
		if (!Arrays.equals(fileHeader, XMA2_ID))
		{
			System.out.println("Filetype not Supported XMA2 Only!!");
			return;
		}
		//Find first packet.
		//When First packet found. Start interleaving.
		int packetOffset = 4;
		xmaBuffer.position(packetOffset);
		int numStreams = getStreams(xmaBuffer, packetOffset);
		for (int i=0; i<numStreams; i++)
		{
			boolean eofReached = false;
			int skipCount = 0;
			String outName = xmaFile.getAbsolutePath().concat("_channel_"+(i+1)+".xma");
			FileOutputStream xmaOutStream = new FileOutputStream(outName);
			while(!eofReached)
			{
				//Extract each one.
				//Read in a packet.
				byte[] nextPacket = new byte[2048];
				if (xmaBuffer.remaining() < 2048)
				{
					break;
				}
				xmaBuffer.get(nextPacket);
				if (skipCount == 0)
				{
					xmaOutStream.write(nextPacket);
					//reset skipcount
					int byteSkipCount = xmaBuffer.array()[xmaBuffer.position()+3-2048];
					if (byteSkipCount < 0) byteSkipCount+=256;
					skipCount = byteSkipCount;
				}
				else
					skipCount--;
				if (!xmaBuffer.hasRemaining())
				{
					eofReached = true;
				}
				
			}
			xmaOutStream.close();
			xmaBuffer.position(packetOffset+2048*(i+1));
			
		}
		//File number of channels.
		*/
		
	}
	
	private static int getStreams(ByteBuffer xmaBuffer, int packetOffset) {
		int streamRepeat = 1; //the lowest possible streamRepeat will happen on Stream2.
		int curStream = 1;
		while(true)
		{
			int frameSkipNum;
			byte frameSkip = xmaBuffer.array()[packetOffset+3+(2048*(curStream-1))];
			frameSkipNum = frameSkip;
			if (frameSkip < 0) frameSkipNum += 256;
			if ((frameSkipNum+curStream < streamRepeat) || (curStream == 1)) streamRepeat = frameSkipNum+curStream;
			if (streamRepeat == curStream ) //This stream is a repeated stream. We must exit.
				return curStream;
			curStream++;
		}
	}

	private static void printUsage()
	{
		System.out.println("Usage...\n");
		System.out.println("java xmaextract xmafile.xma");
		System.out.println("Outputs seperate xma tracks");
	}
}
